/* trt/tom/trt.h.  Generated automatically by configure.  */
/* -*- c -*-
   Written by Pieter J. Schoenmakers <tiggr@gerbil.org>

   Copyright (C) 1996-1999 Pieter J. Schoenmakers.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: trt.h.in,v 1.91 1999/10/05 19:35:58 tiggr Exp $  */

#ifndef _trt_h_include_once
#define _trt_h_include_once

/* The bucket size of the method dispatch tables' buckets.  */
#define TRT_BUCKET_SHIFT	4
#define TRT_BUCKET_SIZE		(1 << TRT_BUCKET_SHIFT)

/* All the rest of this file, the assembler does not understand.  */
#ifndef __ASSEMBLER__

#include <setjmp.h>

#ifndef TRT_INLINE
#define TRT_INLINE static __inline__
#endif

/* For every TOM type, the corresponding type in the C output.  This is
   patched by configure.  */
#define C_BYTE unsigned char
#define C_CHAR unsigned short
#define C_INT int
#define C_LONG long long
#define C_FLOAT float
#define C_DOUBLE double

/* The return type of a function that returns using __builtin_return.  */
#define GENERIC_RETURN_TYPE  void

typedef C_BYTE tom_byte;
typedef C_CHAR tom_char;
typedef C_INT tom_int;
typedef C_LONG tom_long;
typedef C_FLOAT tom_float;
typedef C_DOUBLE tom_double;

/* An integer type large enough to hold all bits in a pointer.  */
#define POINTER_INT_TYPE	C_INT

/* Return a hash value for the pointer P.  There is no need to mix or move
   bits, since the hashtables know how to cope with insignificant bits at
   significant positions.  */
#define HASH_POINTER(P) ((POINTER_INT_TYPE) (P))

/* The CURRENT_UNIT should be set if compilation and resolution is run
   with the -1 option.  */
#ifdef CURRENT_UNIT
#define XSEL(X, Y) (&_sd_ ## X ## _ ## Y)
#define SEL(Y) USEL (CURRENT_UNIT, Y)
#define USEL(X, Y)  XSEL (X, Y)
#else
#define XSEL(Y) (&_sd_ ## Y)
#define SEL(Y) XSEL (Y)
#define USEL(X, Y) XSEL (Y)
#endif

/* The default lookup method for this machine.  It can be overriden (for
   instance on hppa-hpux where LOOKUP_SEND can not be used from within
   dynamically loaded code).  */
#ifndef LOOKUP_DEFAULT
#define LOOKUP_DEFAULT  LOOKUP_LOOKUP
#endif
#define LOOKUP_DIRECT  1
#define LOOKUP_LOOKUP  2
#define LOOKUP_SEND    3

/* Select proper definition of the TRT_SEND macro based on the lookup
   method.  */
#if LOOKUP_DEFAULT == LOOKUP_SEND
#define TRT_SEND(TYPE, RCV, SEL, ARGS...)  \
  (TYPE trt_send) ((RCV), (SEL) , ## ARGS )
#elif LOOKUP_DEFAULT == LOOKUP_LOOKUP
#define TRT_SEND(TYPE, RCV, SEL, ARGS...)  \
  (TYPE (trt_lookup ((struct trt_instance *) (RCV), (SEL))))	\
    ((RCV), (SEL) , ## ARGS )
#else
#error you do not want direct lookups
#endif

/* Which errno to use and how to report errors?  */
#define OS_ERRNO  errno
#define OS_ERROR_MSG  strerror (OS_ERRNO)

/* Names.  A name is a string and its length.  */
struct name
{
  /* The actual string holding the name.  */
  char *s;

  /* The string's length.  */
  int len;
};

typedef struct trtd_selector
{
  /* The unique numeric identification of this selector in the closed
     enumeration of all selectors.  */
  unsigned int sel_id;

  /* The name of the selector.  */
  struct name name;

  /* The description of the arguments (IN) and return values (OUT) of this
     selector.  */
  struct trtd_selector_args *in, *out;
} *selector;

/* Closed enumeration of argument types.  Void is actually never used (as
   its presence only indicates the absence of arguments, i.e. the argument
   count being 0; it can't indicate the type of one of them).  */
enum trt_type_encoding
{
  TRT_TE_VOID,
  TRT_TE_BOOLEAN,
  TRT_TE_BYTE,
  TRT_TE_CHAR,
  TRT_TE_INT,
  TRT_TE_LONG,
  TRT_TE_FLOAT,
  TRT_TE_DOUBLE,
  TRT_TE_POINTER,
  TRT_TE_SELECTOR,
  TRT_TE_REFERENCE,
  TRT_TE_DYNAMIC,

  TRT_TE_NUM
};

/* Struct describing in or out arguments to methods implementing a certain
   selector.  Tuples are flattened before inclusion.  These structures can
   be shared between selectors.  In fact, the selector `int foo' shares
   its OUT arguments description with the IN description of the selectors
   for the methods `void foo int' and `void bar (int)'. */
struct trtd_selector_args
{
  /* The number of entries in ARGS.  This is 0 for void args (i.e. a
     selector returning void or without any arguments).  */
  int num;

  /* The actual descriptions.
     XXX These should occupy a byte each...
     Mon Jul  1 23:00:33 1996, tiggr@tricky.es.ele.tue.nl  */
  enum trt_type_encoding args[0];
};

/* The basic structure of _every_ object (including classes).  */
typedef struct trt_instance
{
  /* The class of this object.  */
  struct trt_class *isa;

  /* The flags needed by the runtime.  */
  tom_int asi;
} *tom_object;

/* Accessors for the ASI of objects.  */

/* Return or set the garbage collection colour of the object.  */
#define TGC_COLOUR(ASI)		(enum tgc_colour) ((ASI) & 0xff)
#define TGC_SET_COLOUR(ASI, C)  ((ASI) = ((ASI) & ~0xff) | (C))

/* Return !0 for objects which are sent a `gc_mark_elements' by the
   garbage collector.  Such objects are objects with a pointer type
   instance variable.  */
#define TGC_EXPLICIT_MARK_P(ASI)  ((ASI) & TGC_ASI_EXPLICIT_MARK_P)
#define TGC_ASI_EXPLICIT_MARK_P	0x00000100

/* Return !0 for containers.  */
#define TGC_CONTAINER_P(ASI)	((ASI) & TGC_ASI_CONTAINER_P)
#define TGC_ASI_CONTAINER_P	0x00000200

/* Return !0 for objects that require stack notify  */
#define TGC_STACK_NOTIFY_P(ASI) ((ASI) & TGC_ASI_STACK_NOTIFY_P)
#define TGC_ASI_STACK_NOTIFY_P  0x00000400

/* Return 1 for class objects.  This excludes instances and meta-class
   objects.  */
#define TGC_CLASS_P(ASI)	((ASI) & TGC_ASI_CLASS_P)
#define TGC_ASI_CLASS_P		0x01000000

/* Return 1 for meta-class objects.  This excludes normal class objects
   and instances.  */
#define TGC_META_CLASS_P(ASI)	((ASI) & TGC_ASI_META_CLASS_P)
#define TGC_ASI_META_CLASS_P	0x02000000

/* Colour values for the garbage collector.  Gray objects are alive and
   must still be scanned.  Black objects are alive and scanned.  White
   objects could be dead, but they could also turn gray.  Pink is a fake
   colour used to count unallocated objects.  Objects are allocated gray.  */
enum tgc_colour
{
  TGC_GRAY,
  TGC_WHITE,
  TGC_BLACK,
  TGC_PINK,
  TGC_NUM,
};

typedef void (*void_imp) ();
typedef tom_byte (*boolean_imp) ();
typedef tom_byte (*byte_imp) ();
typedef tom_char (*char_imp) ();
typedef tom_int (*int_imp) ();
typedef tom_long (*long_imp) ();
typedef tom_float (*float_imp) ();
typedef tom_double (*double_imp) ();
typedef void * (*pointer_imp) ();
typedef selector (*selector_imp) ();
typedef tom_object (*reference_imp) ();

/* The information on each class.  */
struct trt_class_info
{
  /* The method dispatch table (obviously) of our instances.  */
  struct trt_method_dispatch_table *mdt;

  /* The extension offset table (obviously) of our instances.  */
  struct trt_extension_offset_table *eot;

  /* The number of allocated (and currently live) instances of this meta.  */
  unsigned int num_instances;

  /* The name of this class.  */
  struct name name;

  /* The table of offsets to object variables within an instance of this
     class, which have a reference type.  Needed by the garbage collector
     when scanning gray objects.  */
  struct trt_refvar_offset_table *rvo;

  /* The size of instances of this class.  */
  int instance_size;

  /* The value of the ASI for newly allocated instances.  */
  int initial_asi;

  /* The address of our meta-reference, i.e. the pointer to us for the
     purpose of class messaging.  */
  struct trt_class **class_reference;

  /* Description of every extension of instances of this class.  */
  struct trt_meta_extensions *extensions;

  /* The direct submetas and supermetas of this meta.  */
  struct trt_metas *subs, *supers;

  /* The class posing as this class.  */
  struct trt_class *poser;

  /* State carrying extensions contained in instances of this class (or
     the class of this meta).  Created upon request, cleared upon dload.  */
  tom_object state_extensions;

  /* The mark to be used when traversing the class hierarchy.  */
  int mark;

  /* Secondary mark to use when the first mark is already in use.  */
  int mark2;
};

/* The basic structure of class objects and meta-class objects.  The first
   fields _must_ match the fields of trt_instance; the other fields _must_
   match the fields defined for TRO_META_COMMON_INFORMATION.  */
struct trt_class
{
  /* The class of this object.  */
  struct trt_class *isa;

  /* The flags needed by the runtime.  */
  tom_int asi;

  /* The actual information on this class.  */
  struct trt_class_info info;
};

#ifdef TESLA_OUTPUT
/* Redeclaration so tesla output does not need to include tom/util.h.
   Temporary!  Tue Jul 21 13:57:04 1998, tiggr@gerbil.org  */
tom_object byte_string_with_c_string (char *s);
#endif

struct stalloc
{
  /* The number of entries following this struct, and the capacity.  Iff
     CAP == 0, this table is allocated within the text or data segment of
     the program, and should not be freed.  */
  int num, cap;
};

struct trt_method_dispatch_table
{
  struct stalloc st;

  /* Buckets in this table.  */
  struct trt_method_dispatch_bucket *b[0];
};

struct trt_method_dispatch_bucket
{
  /* Methods in this bucket.  */
  int_imp m[TRT_BUCKET_SIZE];

  /* Iff !0, this bucket is allocated within the text or data segment of
     the program, and should not be freed.  */
  unsigned int static_p;

  /* Reference count to this bucket.  */
  int ref_count;

  /* The `owner' of this bucket.  */
  struct trt_method_dispatch_table *owner;
};

struct trt_extension_offset_table
{
  struct stalloc st;

  /* Offsets for the extensions in this table.  */
  unsigned int offset[0];
};

struct trt_refvar_offset_table
{
  struct stalloc st;

  /* Offsets for the reference variables in this table's meta's instances.
     An offset of 0, if still generated by the compiler, can be safely
     ignored, as all classes and meta-classes (which together are the set
     of possible values of ISA) are the root objects for the garbage
     collector.  */
  unsigned int offset[0];
};

/* Description of the extensions of a class or meta-class.  */
struct trt_meta_extensions
{
  struct stalloc st;

  /* The actual descriptions of this meta's extensions.  */
  struct trtd_extension *extensions[0];
};

/* List of related classes.  */
struct trt_metas
{
  struct stalloc st;

  /* The actual metas.  */
  struct trt_class *metas[0];
};

struct trtd_selectors
{
  struct stalloc st;

  /* The actual selectors.  */
  selector selectors[0];
};

/* All registered selectors, ordered and indexable by their identity.  */
extern struct trtd_selectors *trt_selectors;

/* Information common to variable descriptions.  */
struct trtd_var
{
  /* The name of the variable.  */
  struct name name;

  /* The variable's type.  */
  enum trt_type_encoding type;
};

/* Description of a `normal' (i.e. non-static) variable.  */
struct trtd_normal_var
{
  /* Common information.  */
  struct trtd_var c;

  /* This variable's offset from the start of the containing extension.  */
  unsigned int offset;
};

/* Description of a block, of its variables to be more precise.  */
struct trtd_block
{
  /* The number of block variables in this block.  */
  int num_vars;

  /* The descriptions of the block variables.  */
  struct trtd_normal_var vars[0];
};

/* Description of a static variable.  */
struct trtd_static_var
{
  /* Common information.  */
  struct trtd_var c;

  /* This variable's address.  */
  void *address;

  /* Iff !0, this static var is thread-local, and the ADDRESS actually
     points to the integer holding the offset from the start of the
     thread-local data to the value of the variable.  */
  unsigned int th_local: 1;
};

struct trtd_extension
{
  /* The class object describing the object of which we're an extension.  */
  struct trt_class *meta;  

  /* Our {Extension} object, if we already have one.  */
  tom_object extension_object;

  /* The address of the global variable holding our EID.  */
  int *eid_in_a_global;

  /* The name of this extension.  */
  struct name name;

  /* The description of the methods introduced.  */
  struct trtd_methods *methods;

  /* Description of the metas inherited because of this extension.  */
  struct trt_metas *supers;

  /* Size of the state carried by this extension.  */
  int state_size;

  /* The alignment (in bits) required by this extension's variables.  */
  int state_align;

  /* Pointer to the static variables' description, and their number.  */
  struct trtd_static_var *statics;
  int num_statics;

  /* The number of non-static variables in this extension, i.e. the number
     of entries in VARS.  Iff NUM_VARS == 0, EID_IN_A_GLOBAL is NULL.  */
  int num_vars;

  /* The descriptions of the non-static variables.  */
  struct trtd_normal_var vars[0];
};

/* Unit description.  */
struct trtd_unit
{
  /* This unit's name.  */
  struct name name;

  /* The extensions defined in this unit, and their number.  These
     extensions are are ordered according to a topological sort of the
     inheritance graph.  */
  struct trtd_extension **extensions;
  int num_extensions;

  /* The classes defined in this unit, and their number.  */
  struct trt_class **classes;
  int num_classes;
};

struct trtd_selector_map
{
  /* The actual selector, and the dynamic selector onto which it should be
     mapped.  */
  selector actual, dynamic;
};

/* Recorded dynamic selectors.  */
struct trt_dynamic_selectors
{
  struct stalloc st;

  struct trtd_selector_map *maps[0];
};

/* Module description.  */
struct trtd_module
{
  /* The units in this module, and their number.  */
  struct trtd_unit **units;
  int num_units;

  /* The selectors used by this module.  */
  struct trtd_selectors *selectors;

  /* The dynamic selectors introduced by this module.  */
  struct trtd_selector_map *dynamic_selectors;
  int num_dynamic_selectors;

  /* The super references in this module, and their number.  */
  struct trtd_super *super_refs;
  int num_super_refs;

  /* The load methods defined in this module, and their number.  */
  struct trtd_load_imp *load_imps;
  int num_load_imps;

  /* The selector of the main method defined by this module, and the address
     of the pointer to the class to receive it.  */
  selector main_selector;
  struct trt_class **main_class;

  /* The poser/posed relationships introduced by this module.  */
  struct trtd_poser *posers;
  int num_posers;
};

/* Description of a reference to `super'.  */
struct trtd_super
{
  /* The address of this super ref.  */
  struct trt_class **ref;

  /* The referencing class and the super class being referenced.  */
  struct trt_class **sub, **super;
};

/* Description of a posing relationship.  */
struct trtd_poser
{
  struct trt_class *poser, *posed;
};

struct trtd_method
{
  /* The method implementation.  */
  int_imp imp;

  /* The meta providing this implementation.  */
  struct trt_class *meta;

  /* The corresponding selector.  */
  selector sel;
};

struct trtd_methods
{
  /* The number of methods.  */
  int num_methods;

  /* The actual methods.  */
  struct trtd_method methods[0];
};

/* Description of a load implementation, i.e. an implementation by a class
   extension of the `void load' method.  */
struct trtd_load_imp
{
  /* The implementation.  */
  int_imp imp;

  /* The address of a pointer to the class object to be used as the
     receiver.  */
  struct trt_class **class;
};

/* Return a pointer to the implementation for the receiver RCV of the
   selector SEL.  */
int_imp trt_lookup (tom_object rcv, selector sel);

/* Return a pointer to the implementation of the selector SEL as if the
   (imaginary) receiver had the class SUPER.  */
int_imp trt_lookup_super (struct trt_class *super, selector sel);

/* Perform a lookup of the selector SEL for the receiver RCV and invoke
   that method with the same arguments.  This might not be available on
   all machines.  */
tom_int trt_send (void *rcv, selector sel, ...);

/* Stack protection struct for machines which (not yet) have their stack
   scanned automatically.  */
struct trt_protect_stack
{
  /* The next entry up the stack.  */
  struct trt_protect_stack *next;

  /* The number of locals, the address of which follows this struct, and
     the number of arguments.  */
  int num_locals, num_args;

  void ***arguments;
};

/* The stack protection stack.  */
extern struct trt_protect_stack *_trt_sp;

/* Colour the OBJECT gray and put it on the gray list.  The object should
   be white.  */
void trt_make_gray (struct trt_instance *object);

/* The OBJECT must not be white.  If it is, it will be grayed.  */
TRT_INLINE void *
trt_assign_local_var (void *object)
{
  struct trt_instance *o = object;

  if (o && TGC_COLOUR (o->asi) == TGC_WHITE)
    trt_make_gray (o);

  return o;
}

/* If the OBJECT of which some object variable will be set to the VALUE,
   is black, make the VALUE object be at least gray.  */
TRT_INLINE void *
trt_assign_object_var (void *object, void *value)
{
  struct trt_instance *o = object;
  struct trt_instance *v = value;

  if (v && TGC_COLOUR (v->asi) == TGC_WHITE
      && TGC_COLOUR (o->asi) == TGC_BLACK)
    trt_make_gray (v);

  return v;
}

/* Information common to the non-local exit mechanism.  Providing yet
   another C-implemented ad-hoc dynamic dispatch paradigm.  */
struct trt_condition_info
{
  /* Indication of what follows this struct: information for a bind,
     catch, or unwind.  */
  enum
  {
    TRT_CONDITION_KIND_BIND,
    TRT_CONDITION_KIND_CATCH,
    TRT_CONDITION_KIND_UNWIND,
  } kind;
};

/* Struct to register the handlers of a bind.  */
struct trt_bind
{
  /* Common information.  */
  struct trt_condition_info c;

  /* A pointer to the handler (C) function, as generated by tomc.  The first
     argument is the index into CLASSES of the class matching the class of
     the CONDITION being raised or signaled.  (There is only one handler to
     reduce the overhead of taking the address of a nested function.)  */
  void * (*tomc_handler) (int class, void *condition);

  /* The number of entries in HANDLERS.  */
  int num;

  /* The condition classes of interest, with their handlers.  */
  struct
  {
    void *condition_class, *condition_handler; 
  } handlers[0];
};

/* Struct to register a catch.  */
struct trt_catch
{
  /* Common information.  */
  struct trt_condition_info c;

  /* The object being the tag which can be thrown.  */
  void *tag;

  /* The jmp buffer to which to longjmp if this catch catches a throw.  */
  jmp_buf *jmp;

  /* Iff !0, the value to be assigned to _TRT_SP upon taking this jump.  */
  void *sp;

  /* Description of the throw value expected by this catch.  */
  struct trtd_selector_args *value_desc;

  /* For each `return' value, a pointer to where the value is to be put.  */
  void *values[0];
};

/* Description of an unwind protection.  */
struct trt_unwind
{
  /* Common information.  */
  struct trt_condition_info c;

  /* The protection (C) function to be invoked during an unwind.  */
  void (*protect) (void);
};

/* Functions to register or unregister one of the above bind, catch, or
   unwind structs.  */
void trt_register (struct trt_condition_info *reg);
void trt_unwind (struct trt_condition_info *reg);

/* Other functions used by tomc compiled code.  */

/* Return a pointer to the local data of this thread.  */
void *trt_thread_get_local_data (void);

/* Instantiate a new block with the indicated BODY of code and the TRIGGER
   describing the argument and return types.  CONTEXT is the context in
   which the block is created.  VARS points to the block variables, which
   are described by DESC.  */
void *trt_new_block (void (*body) (), selector trigger, void *context,
		     void *variables, struct trtd_block *desc);

#endif
#endif
